using Microsoft.AspNetCore.SignalR;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using System.Collections.Concurrent;

namespace SignalRChat.Hubs
{
    public class Overlay
    {
        public int X { get; set; }
        public int Y { get; set; }
        public string ID { get; set; }
        public bool Intersection { get; set; }
        public int Rotation { get; set; }
        public override string ToString()
        {
            return $"{X}_{Y}_{Intersection}";
        }
    }

    public class Tile
    {
        public int X { get; set; }
        public int Y { get; set; }
        public string ID { get; set; }
        public int Rotation { get; set; }


        public override string ToString()
        {
            return $"{X}_{Y}";
        }
    }

    public class Wall
    {
        public int X { get; set; }
        public int Y { get; set; }
        public string ID { get; set; }
        public int Rotation { get; set; }
        public override string ToString()
        {
            return $"{X}_{Y}";
        }
    }

    public class Player
    {
        public string ConnectionId { get; set; }
        public string Color { get; set; }
        public string Initials { get; set; }
        public Tile Position { get; set; }
    }

    public class SessionData

    {
        public SessionData()
        {
            Map = new ConcurrentDictionary<string, Tile>();
            Overlays = new ConcurrentDictionary<string, Overlay>();
            Walls = new ConcurrentDictionary<string, Wall>();
            Players = new List<Player>();
            Colors = new List<string>();
        }
        public ConcurrentDictionary<string, Tile> Map { get; set; }
        public ConcurrentDictionary<string, Wall> Walls { get; set; }
        public ConcurrentDictionary<string, Overlay> Overlays { get; set; }
        public bool IsValid { get; set; }
        public List<Player> Players { get; set; }
        public List<string> Colors;
    }

    public class ChatHub : Hub
    {
        private static Dictionary<string, SessionData> _sessions = new Dictionary<string, SessionData>();
        public List<string> Colors = new List<string>{"CC0000",
"CC3300",
"FFCC00",
"009900",
"006666",
"0066FF",
"0000CC",
"663399",
"CC0099"};

        public async Task SendMessage(string user, string message, string sessionName)
        {
            await Clients.Group(sessionName).SendAsync("ReceiveMessage", user, message);
        }

        public async Task NewTile(string sessionName, Tile tile)
        {
            var existingTile = _sessions[sessionName].Map.TryGetValue(tile.ToString(), out var t);
            if (existingTile)
            {

                _sessions[sessionName].Map.TryRemove(t.ToString(), out var rtile);
            }
            _sessions[sessionName].Map.TryAdd(tile.ToString(), tile);


            await Clients.Group(sessionName).SendAsync("NewTile", tile);
        }

        public async Task NewWall(string sessionName, Wall tile)
        {
            var existingTile = _sessions[sessionName].Walls.TryGetValue(tile.ToString(), out var t);
            if (existingTile)
            {
                _sessions[sessionName].Walls.TryRemove(t.ToString(), out var rtile);

            }
            _sessions[sessionName].Walls.TryAdd(tile.ToString(), tile);
            await Clients.Group(sessionName).SendAsync("NewWall", tile);
        }

        public async Task NewOverlay(string sessionName, Overlay tile)
        {
            var existingTile = _sessions[sessionName].Overlays.TryGetValue(tile.ToString(), out var t);
            if (existingTile)
            {
                _sessions[sessionName].Overlays.TryRemove(t.ToString(), out var rtile);

            }
            _sessions[sessionName].Overlays.TryAdd(tile.ToString(), tile);
            await Clients.Group(sessionName).SendAsync("NewOverlay", tile);
        }

        public async Task DeleteTile(string sessionName, Tile tile)
        {
            _sessions[sessionName].Map.TryRemove(tile.ToString(), out var rtile);


            await Clients.Group(sessionName).SendAsync("DeleteTile", tile);
        }

        public async Task DeleteWall(string sessionName, Wall tile)
        {
            _sessions[sessionName].Walls.TryRemove(tile.ToString(), out var rtile);


            await Clients.Group(sessionName).SendAsync("DeleteWall", tile);
        }

        public async Task DeleteOverlay(string sessionName, Overlay tile)
        {
            _sessions[sessionName].Overlays.TryRemove(tile.ToString(), out var rtile);


            await Clients.Group(sessionName).SendAsync("DeleteOverlay", tile);
        }

        public async Task<SessionData> NewSession(string sessionName, string initials)
        {
            var session = new SessionData();
            session.Colors = new List<string>(Colors);
            session.Colors.Shuffle();
            var player = new Player { Position = new Tile { X = 0, Y = 0 }, ConnectionId = Context.ConnectionId, Color = session.Colors[0], Initials = initials };
            session.Players.Add(player);
            _sessions.Add(sessionName, session);

            await Groups.AddToGroupAsync(Context.ConnectionId, sessionName);
            return session;
        }

        public async Task<SessionData> JoinSession(string sessionName, string initials)
        {
            if (_sessions.ContainsKey(sessionName))
            {
                var session = _sessions[sessionName];
                var player = new Player { Position = new Tile { X = 0, Y = 0 }, ConnectionId = Context.ConnectionId, Color = session.Colors[session.Players.Count], Initials = initials };
                session.Players.Add(player);
                await Clients.Group(sessionName).SendAsync("NewPlayer", player);
                await Groups.AddToGroupAsync(Context.ConnectionId, sessionName);
                return session;
            }
            else
            {
                return null;
            }
        }

        public async Task UpdatePosition(string sessionName, Tile tile)
        {
            var player = _sessions[sessionName].Players.First(m => m.ConnectionId == Context.ConnectionId);
            player.Position = tile;
            await Clients.Group(sessionName).SendAsync("PlayerUpdate", player);

        }

        public async Task<SessionData> Refresh(string sessionName)
        {
            return _sessions[sessionName];
        }

        public async Task Sync(string sessionName, SessionData map)
        {
            _sessions[sessionName].Map = map.Map;
            _sessions[sessionName].Overlays = map.Overlays;
            _sessions[sessionName].Walls = map.Walls;
            
            await Clients.Group(sessionName).SendAsync("UpdateMap",  _sessions[sessionName]);

        }
    }

    public static class ExtensionMethods
    {
        private static Random rng = new Random();

        public static void Shuffle<T>(this IList<T> list)
        {
            int n = list.Count;
            while (n > 1)
            {
                n--;
                int k = rng.Next(n + 1);
                T value = list[k];
                list[k] = list[n];
                list[n] = value;
            }
        }
    }
}